/****************************************************/
/*                                                  */
/* hw_SH2A.c : SH-2A/FPU specific routines          */
/*                                                  */
/* NON COMMERCIAL USE ONLY                          */
/*                                                  */
/* 2010/05/12 Rev. 1.01 by pcm1723                  */
/*                                                  */
/****************************************************/


#include "iodefine.h"
#include <machine.h>
#include "../common/tonegen.h"

// SSIF clock divider and system word length selection
//
// CKDV   
//    0 : 1/1
//    1 : 1/2
//    2 : 1/4
//    3 : 1/8
//    4 : 1/16
//
// SWL
//    2 : 24 bit system word
//    3 : 32 bit system word
//     
struct {
   unsigned char ckdv; // ClocK Divisor
   unsigned char swl;  // System Word Length
} CKWL_tab[13] = {
    {0, 3}, // 
    {0, 3}, //  64 fs =  1 * 32 * 2
    {1, 3}, // 128 fs =  2 * 32 * 2 
    {2, 2}, // 192 fs =  4 * 24 * 2
    {2, 3}, // 256 fs =  4 * 32 * 2
    {2, 3}, // 320 fs 
    {3, 2}, // 384 fs =  8 * 24 * 2
    {3, 2}, // 448 fs
    {3, 3}, // 512 fs =  8 * 32 * 2
    {4, 2}, // 576 fs
    {4, 2}, // 640 fs
    {4, 2}, // 704 fs
    {4, 2}  // 768 fs = 16 * 24 * 2
}; // struct {} CKWL_tab[]

// 
// set configuration values for specified mode 
//
void set_config(int mode)
{
//  T0MR0      = sys_config[fastchip][mode].t1div - 1; // set Timer divisor constant
  MAXUNITS   = sys_config[fastchip][mode].polyphony; // polyphony = number of units
  SEQTICK1ms = sys_config[fastchip][mode].tick1ms;   // tick count for 1ms
  SEQTICK4ms = sys_config[fastchip][mode].tick4ms;   // tick count for 4ms
  oct_ofs    = sys_config[fastchip][mode].oct_ofs;   // octave offset
  note_ofs   = sys_config[fastchip][mode].note_ofs;  // f-number table offset
  memcpy(&FS_ASC, sys_config[fastchip][mode].fs_asc, 2);    // ASCII code for fs
  
  SRC0.SRCCTRL.BIT.SRCEN = 0; // SRC disable    
  SRC0.SRCCTRL.BIT.IFS        // input fs
      = sys_config[fastchip][mode].t1div;
  SRC0.SRCCTRL.BIT.SRCEN = 1; // SRC enable    
  
#ifdef REVERB
  delay_mem_size = SEQTICK1ms * delay_ms; // calc delay momory size
  if (DELAYMEMSIZE < delay_mem_size) delay_mem_size = DELAYMEMSIZE; // clip if too big
  delayp = 0;
#endif
  midi_mon_part = 0;  // reset MIDI monitor part
  MAXMIDICH  = MAXUNITS;                   // supported MIDI channels
} // void set_config()

void SPDIF_interrupt_handler( void )
{
    volatile uint32_t lw;
    if (1 == SPDIF.STAT.BIT.CBTX) { // SPDIF Tx buffer empty intr ?
        if (1 == SRC0.SRCSTAT.BIT.OINT) { // any SRC output data ?
          lw = SRC0.SRCOD.LONG; // read SRC out FIFO if so
          SRC0.SRCSTAT.BIT.OINT = 0; // clear flag
          SPDIF.TLCA = (lw >> 16); // L ch data
          SPDIF.TRCA = (lw);       // R ch data
        } else {
          lw = 0; // avoid FIFO read to prevent underrun error
        }
        DAC_SSIF.SSIFTDR = lw;   // output to DAC via SSIF
        if (1 == SRC0.SRCSTAT.BIT.IINT) { // SRC input FIFO empty?
          SRC0.SRCSTAT.BIT.IINT = 0; // clear flag
          lw = (0xffff & wave_que[wq_rd_ix]); // get wave data
          lw |= (lw << 16); // copy to upper halfword
          SRC0.SRCID.LONG = lw; // input to SRC
          wq_rd_ix = (wq_wr_ix + 1) & (WAVEQUESIZE-1); // adv index
        } // if
    } // if (1 == SPDIF. ...)
} // void SPDIF_interrupt_handler()

void SSIF_interrupt_handler( void )
{
    volatile uint32_t lw;
    if (1 == DAC_SSIF.SSIFSR.BIT.TDE) { // SSIF Tx Data Empty intr ?
        if (1 == SRC0.SRCSTAT.BIT.OINT) { // any SRC output data ?
          lw = SRC0.SRCOD.LONG; // read SRC out FIFO if so
          SRC0.SRCSTAT.BIT.OINT = 0; // clear flag
#ifdef USE_SPDIF          
          SPDIF.TLCA = (lw >> 16); // L ch data
          SPDIF.TRCA = (lw);       // R ch data
#endif          
        } else {
          lw = 0; // avoid FIFO read to prevent underrun error
        }
        DAC_SSIF.SSIFTDR = lw;   // output to DAC via SSIF
        DAC_SSIF.SSIFSR.BIT.TDE = 0;   // clear Tx Data Empty flag
        if (1 == SRC0.SRCSTAT.BIT.IINT) { // SRC input FIFO empty?
          SRC0.SRCSTAT.BIT.IINT = 0; // clear flag
          lw = (0xffff & wave_que[wq_rd_ix]); // get wave data
          lw |= (lw << 16); // copy to upper halfword
          SRC0.SRCID.LONG = lw; // input to SRC
          wq_rd_ix = (wq_wr_ix + 1) & (WAVEQUESIZE-1); // adv index
        } // if
    } // if (1 == SSIF. ...)
} // void SSIF_interrupt_handler()

void enable_interrupt(void)
{
    set_imask(0); // in <machine.h>
} // void enable_interrupt() 

void SPDIF_setup(int use_intr)
{
    INTC.IPR16.BIT._SPDIF  = 0; // disable SPDIF interrupt
    CPG.STBCR7.BIT.MSTP76  = 0; // enable SPDIF clock
    SPDIF.CTRL.BIT.TME     = 0; // disable SPDIF Tx
    SPDIF.CTRL.BIT.CKS     = USE_AUDIO_CLK; // 0 for AUDIO_X1, 1 for AUDIO_CLK
    SPDIF.CTRL.BIT.AOS     = 1; // audio only sample
// L channel status bits    
    SPDIF.TLCS.BIT.FS      = 2*SRC_OFS; // 0 for 44.1 kHz, 2 for 48 kHz
    SPDIF.TLCS.BIT.CHNO    = 1; // 1 for L ch
    SPDIF.TLCS.BIT.CTL     = 2; // Copy enabled
// R channel status bits
    SPDIF.TRCS.BIT.FS      = 2*SRC_OFS; // 0 for 44.1 kHz, 2 for 48 kHz
    SPDIF.TRCS.BIT.CHNO    = 2; // 1 for R ch
    SPDIF.TRCS.BIT.CTL     = 2; // Copy enabled
    SPDIF.CTRL.BIT.TME     = 1; // enable SPDIF Tx
    PORT.PFCR3.BIT.PF12MD  = 5; // select SPDIF output on PF12
    if (use_intr) {             // use interrupt
      SPDIF.CTRL.BIT.TCBI    = 1; // Xmit Ch Buf Int enable
      INTC.IPR16.BIT._SPDIF  = 1; // enable  SPDIF interrupt
    } else {
      SPDIF.CTRL.BIT.TCBI    = 0; // Xmit Ch Buf Int disable
    } // if (use_intr) ...
    SPDIF.TLCA             = 0x0000;  // clear L ch xmit data
    SPDIF.TRCA             = 0x0000;  // clear R ch xmit data
} // void SPDIF_setup()

void SRC_setup( void )
{
    CPG.STBCR6.BIT.MSTP62  = 0; // enable SRC0 clock
    SRC0.SRCCTRL.BIT.SRCEN = 0; // SRC disable    
    SRC0.SRCCTRL.BIT.IFS   = 6; // input  fs = 12 kHz
    SRC0.SRCCTRL.BIT.OFS   = SRC_OFS; // output fs, 0 for 44.1 kHz, 1 for 48 kHz
    SRC0.SRCCTRL.BIT.SRCEN = 1; // SRC enable    
} // void SRC_setup()

void CPG_setup( void )
{
    int clock_mode = (2 * MD_CLK1 + MD_CLK0); 
    switch (clock_mode) {
      case 0: // clock mode 0
// EXTAL 10~12 MHz, PLL x12, 12:4:2, I=120~144M, B=40~48M, P=20~24M
        CPG.FRQCR.WORD = 0x0104; 
        break;
      case 1: // clock mode 1
// USB 48 MHz clk, PLL x12, 3:1:1/2, I=144M, B=48M, P=24M
        CPG.FRQCR.WORD = 0x0104; 
        break;
      case 2: // clock mode 2
// EXTAL 10~18 MHz, PLL x8, 8:4:2, I=120~144M, B=40~72M, P=20~36M
        CPG.FRQCR.WORD = 0x0003; 
        break;
      case 3: // clock mode 3
// USB 48 MHz clk, PLL x8, 8/3:4/3:2/3, I=128M, B=64M, P=32M
        CPG.FRQCR.WORD = 0x0003; 
        break;
    } // switch (clock_mode) ...
} // void CPG_setup()

void GPIO_setup( void )
{
// LED port setup    
	PORT.PCDR0.BIT.PC8DR   = 1; // PC8 data = "1"
	PORT.PCIOR0.BIT.PC8IOR = 1; // PC8 dirction = output
	PORT.PCCR2.BIT.PC8MD   = 0; // PC8 mode = GPIO
#ifdef FORCE_SERIAL_MIDI
    serial_midi = 1; // force serial MIDI
#else    
#ifdef FORCE_LEGACY_MIDI
    serial_midi = 0; // force (legacy) MIDI
#else
// serial MIDI sens port setup
	PORT.PJCR0.BIT.PJ3MD   = 0; // PJ3 mode = GPIO
	PORT.PJIOR0.BIT.PJ3IOR = 0; // PJ3 direction = input
	serial_midi = PORT.PJPR0.BIT.PJ3PR; // read PJ3
#endif
#endif
// AUDIO_CLK input setup (if any)
#if (1 == USE_AUDIO_CLK)
	PORT.PECR0.BIT.PE0MD = 2; // PE0 = AUDIO_CLK
#endif   
// SPDIF_EN signal
#ifdef USE_SPDIF_EN
	PORT.PJCR0.BIT.PJ1MD   = 0; // PJ1 mode = GPIO
	PORT.PJIOR0.BIT.PJ1IOR = 1; // PJ1 direction = output
	PORT.PJDR0.BIT.PJ1DR   = 1; // PJ1 data = "1"
#endif
#ifdef SH2A_DEBUG
	PORT.PCCR1.BIT.PC7MD   = 2; // PC7 mode = TIOC4C
	PORT.PCIOR0.BIT.PC7IOR = 1; // TIOC4C direction = output
#endif
} // void GPIO_setup()

#ifdef SH2A_DEBUG
void MTU_setup()
{
    CPG.STBCR3.BIT.MSTP35 = 0; // enable MTU2 clock
    MTU2.TCR_4.BIT.CCLR  = 5; // clear TCNT on TGRC compare match
    MTU2.TCR_4.BIT.CKEG  = 0; // count on rising edge
    MTU2.TCR_4.BIT.TPSC  = 1; // clock = P_phi / 4
    MTU2.TOER.BIT.OE4C   = 1; // enable TIOC4C outut
    MTU2.TMDR_4.BIT.BFA  = 0; // no buffered TGRA_4/TGRC_4
    MTU2.TMDR_4.BIT.MD   = 0; // PWM mode 1
    MTU2.TIORL_4.BIT.IOC = 3; // TIOC4C toggle on TGRC_4 compare match
    MTU2.TGRC_4.WORD     = 125-1;// for period = 125
    MTU2.TSTR.BIT.CST4   = 1; // start timer 4
} // void MTU_setup()
#endif

void SCIF_setup( void )
{
    CPG.STBCR4.BIT.MSTP45  = 0; // enable SCIF2 clock
    
    MIDI_SCIF.SCSCR.WORD = 0x0000; // Rx, Tx disabled
    MIDI_SCIF.SCFCR.BIT.TFRST = 1; // reset Tx FIFO
    MIDI_SCIF.SCFCR.BIT.RFRST = 1; // reset Rx FIFO
    if (MIDI_SCIF.SCFSR.WORD) {}   // dummy read
    MIDI_SCIF.SCFSR.WORD = 0x0000; // reset status flags
    if (MIDI_SCIF.SCLSR.WORD) {}   // dummy read
    MIDI_SCIF.SCLSR.WORD = 0x0000; // reset overrun flag
    MIDI_SCIF.SCSMR.WORD = 0x0000; // 8N1, Baud clk = P_phi
    MIDI_SCIF.SCEMR.WORD = 0x0080; // Double speed BRG
    if (serial_midi) { // set Baud rate to 38.4 kbps
      MIDI_SCIF.SCBRR.BYTE = 38;     // 24 MHz/(38+1)/16=38.462 kbps (+0.16%)
    } else { // set Baud rate to 31.25 kbps (normal MIDI)
      MIDI_SCIF.SCBRR.BYTE = 47;     // 24 MHz/(47+1)/16=31.25 kbps (+-0%)
    } // if (serial_midi) ...
    MIDI_SCIF.SCFCR.WORD = 0x0000; // SCIF FIFO Control
    
    PORT.PFCR0.BIT.PF2MD   = 4; // enable TxD2 on PF2
    PORT.PFCR0.BIT.PF1MD   = 4; // enable RxD2 on PF1
    
    MIDI_SCIF.SCSCR.WORD = 0x0018; // ASYNC, Rx, Tx enabled, No interr
} // void SC_setup()

void SSIF_setup(int use_intr)
{
    int i;
    CPG.STBCR6.BIT.MSTP65   = 0; // enable SSIF2 clock
    PORT.PFCR1.BIT.PF5MD    = 2; // enable SSIDATA2 on PF5
    PORT.PFCR1.BIT.PF4MD    = 2; // enable SSIWS2   on PF4
    PORT.PFCR0.BIT.PF3MD    = 2; // enable SSICLK2  on PF3
    
    DAC_SSIF.SSICR.LONG     = 0; // reset Control Register
    DAC_SSIF.SSICR.BIT.CKS  = USE_AUDIO_CLK; // 0 for AUDIO_X1, 1 for AUDIO_CLK
    DAC_SSIF.SSICR.BIT.CHNL = 0; // CHaNnel(s) per system-word = 1
    DAC_SSIF.SSICR.BIT.DWL  = 1; // Data   Word Length = 16 bit
// set System Work Length to 32 bit or 24 bit
    DAC_SSIF.SSICR.BIT.SWL  = CKWL_tab[OVS_FACTOR/64].swl;
    DAC_SSIF.SSICR.BIT.SCKD = 1; // Serial CloCK Direction = OUT
    DAC_SSIF.SSICR.BIT.SWSD = 1; // Serial Word Select Direction = OUT
    DAC_SSIF.SSICR.BIT.SCKP = 0; // DATA/WS change on falling edge of CLK
    DAC_SSIF.SSICR.BIT.SWSP = 1; // Serial Word Select Polarity: L = 1
    DAC_SSIF.SSICR.BIT.SDTA = 1; // Serial DaTa Alignment = padding, data
    DAC_SSIF.SSICR.BIT.PDTA = 1; // Parallel DaTa Alignment = MS-word first
    DAC_SSIF.SSICR.BIT.DEL  = 1; // no serial data delay
// set SSICLK division ratio to 1/1, 1/2, 1/4, 1/8, 1/16
    DAC_SSIF.SSICR.BIT.CKDV = CKWL_tab[OVS_FACTOR/64].ckdv; 
    DAC_SSIF.SSICR.BIT.TEN  = 1; // Transmission ENable
    if (use_intr) { // use interrupt
      DAC_SSIF.SSIFCR.BIT.TIE  = 1; // Tx Interrupt Enable
      INTC.IPR15.BIT._DAC_SSIF  = 1; // enable SSIF interrupt
    } else {
      DAC_SSIF.SSIFCR.BIT.TIE  = 0; // Tx Interrupt Disable
    } // if (use_intr) ...    
    for (i = 0; i < 4; i++) { // fill xmit FIFO halfway
        DAC_SSIF.SSIFTDR = 0; // fill zero
    } // for
} // void SSIF_setup()

void SH2A_port_setup( void )
{   
    CPG_setup();    // Clock Pulse Generator setup
    GPIO_setup();   // General Purpose I/O setup
#ifdef SH2A_DEBUG
    MTU_setup();    // Multifunction Timer pulse Unit setup
#endif    
    SCIF_setup();   // Serial Communication InterFace setup
#ifdef USE_SPDIF    
    SPDIF_setup(0); // SPDIF setup
#endif
    SSIF_setup(1);  // Serial Sound InterFace setup for DAC
    SRC_setup();    // Sampling Rate Converter setup
} // void SH2A_port_setup()
